/*
 * Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 */
import { Component, OnInit, OnDestroy, ViewChild, ViewEncapsulation } from "@angular/core";
import { Title } from "@angular/platform-browser";
import { ActivatedRoute, Router, Params } from "@angular/router";
import { Capability } from "../../model/capability";
import { CapabilityRequest } from "../../model/capability-request";
import { CapabilitiesService } from "../../services/capabilities.service";
import { CapabilityRequestService } from "../../services/capability-request.service";
import { CapabilityLauncherService } from "../../services/capability-launcher.service";
import { DataRetrieverService } from "../../services/data-retriever.service";
import { PollingDataUpdaterService } from "../../services/polling-data-updater.service";
import { BehaviorSubject, combineLatest, Observable, Subject } from "rxjs";
import { map, repeatWhen, scan, takeUntil, startWith } from "rxjs/operators";
import { Staff } from "../../model/staff";
import { ActiveRequestsService } from "../../services/active-requests.service";
import { StorageService } from "../../../shared/storage/storage.service";
import { Filter } from "./components/filter-menu/filter-menu.component";
import { FormControl } from "@angular/forms";
import { MatPaginator, PageEvent } from "@angular/material/paginator";
import { WsHeaderComponent } from "../../ws-header/ws-header.component";
import { RxState } from "@rx-angular/state";

export const defaultSortOrder = "desc";
export const defaultColumn = "id";

/**
 * Encapsulates data necessary for sorting so it can be managed by RxState
 *
 * # Parameters
 * * `sortCol` - A `string` with the column to sort on, e.g. `id`
 * * `sortDir` - A `string` containing the sort direction, either `asc` or `desc`
 */
interface SortState {
  sortCol: string;
  sortDir: {dir: string, col: string};
}

const initSortState = {
  sortCol: "id",
  sortDir: { dir: defaultSortOrder, col: defaultColumn },
};

@Component({
  selector: "app-active-capability-requests",
  templateUrl: "./active-capability-requests.component.html",
  styleUrls: ["./active-capability-requests.component.scss"],
  providers: [RxState],
  encapsulation: ViewEncapsulation.None,
})
export class ActiveCapabilityRequestsComponent implements OnInit, OnDestroy {
  model$: Observable<SortState> = this.state.select();

  public title = "Active Capability Requests";
  public activeCapabilityRequests: Array<CapabilityRequest>;
  public capability: Capability;
  public isPaused: boolean;
  public enabledCapabilities: Array<string>;
  public qaStaff: Array<Staff>;
  public capabilityName: string;
  public stdCalibrationCapability = "std_calibration";
  private activeCapabilityNames: Observable<any>;
  private paramsObj: any;
  private filters: any;
  public showFilterMenu = false;
  private notSubmittedState = "Latest version has not been submitted";
  private newParams: Params;
  private filterSets = {};
  private availableFilters: any;
  // The sortDirection observable gets activated twice when clicking sort, causing it to immediately undo itself, this
  // flag is to prevent this

  // Filter sets
  public statesSet = new Set<string>();
  public execStatusSet = new Set<string>();
  public downloadStatusSet = new Set<string>();
  public stage1QaSet = new Set<string>();
  public stage2QaSet = new Set<string>();
  public srdpSet = new Set<string>();
  public obsDateSet = new Set<string>();

  // Needed for autocomplete
  stage2QaControl: FormControl = new FormControl();
  stage1QaControl: FormControl = new FormControl();
  filteredStage2QaStaff: Observable<Staff[]>;
  filteredStage1QaStaff: Observable<Staff[]>;

  public statesToFilter = [
    { name: "Complete" },
    { name: "Submitted" },
    { name: "Created" },
    { name: "Cancelled" },
    { name: "Failed" },
  ];
  public execStatusToFilter = [
    { name: "Not submitted", filter_val: this.notSubmittedState },
    { name: "Start", filter_val: "Start" },
    { name: "Queued", filter_val: "Queued" },
    { name: "Executing", filter_val: "Executing" },
    { name: "Ingesting", filter_val: "Ingesting" },
    { name: "Awaiting QA", filter_val: "Awaiting QA" },
    { name: "Stage 2 Review", filter_val: "Stage 2 Review" },
    { name: "QA Closed", filter_val: "QA Closed" },
    { name: "Failed", filter_val: "Failed" },
    { name: "Error", filter_val: "Error" },
    { name: "Cancelled", filter_val: "Cancelled" },
    { name: "Complete", filter_val: "Complete" },
  ];
  public downloadStatusToFilter = [
    { name: "Unknown", filter_val: "Unknown" },
    { name: "Not Started", filter_val: "Not Started" },
    { name: "In Progress", filter_val: "In Progress" },
    { name: "Complete", filter_val: "Complete" },
  ];
  public srdpOptions = [{ name: "true" }, { name: "false" }];
  public obsDateOptions = [{ obs_min: "" }, { obs_max: "" }];

  // Observers
  private activeRequestsObserver = {
    next: (request) => {
      // Set active requests and perform sorting
      this.activeCapabilityRequests = request.active_requests;
      this.sortedActiveRequests$ = this.sortDirection$.pipe(
        map((sort) => {
          this.activeCapabilityRequests = this.filterFrontendFilters(this.activeCapabilityRequests);
          this.pageAllVal = this.activeCapabilityRequests.length;
          return !sort.col
            ? this.activeCapabilityRequests
            : this.sortByColumn(this.activeCapabilityRequests, sort.col, sort.dir);
        }),
      );
    },
    error: (error) =>
      console.error("Error when retrieving list of active capability requests:", error),
  };

  public capabilityObserver = {
    next: (capability) => {
      this.capability = capability;
      this.isPaused = this.capability.paused;
    },
    error: (error) => console.error("Error when retrieving capability:", error),
  };

  // Table sorting
  public sortedActiveRequests$: Observable<any>;
  private sortedColumn$ = new BehaviorSubject<string>("");
  public sortCol = defaultColumn;
  public sortDir = this.getSortDirection();

  private sortDirection$ = this.sortedColumn$.pipe(
    scan<string, { col: string; dir: string }>(
      (sort, val) => {
        if (sort.col === val) {
          this.activeRequestsService.saveReqIdSortOrder(this.sortDir === "asc" ? "desc" : "asc");
          this.sortDir = this.getSortDirection();
        }

        return { col: val, dir: this.sortDir };
      },
      { dir: this.sortDir, col: "" },
    ),
  );

  // to signal when all subscribers should unsubscribe (triggered in tear down)
  private ngUnsubscribe = new Subject();

  // Needed for pagination of requests table
  @ViewChild("requestPaginator") requestsPaginator!: MatPaginator;
  pageSize: number;
  pageStart: number;
  pageStop: number;
  pageAllVal: number;
  getPageSizeOptions(): number[] {
    const pageSizes = [20, 40];
    const pageSizeOptions = [this.pageAllVal];

    pageSizes.forEach((size) => {
      if (this.pageAllVal > size) pageSizeOptions.unshift(size);
    });

    return pageSizeOptions;
  }

  public sortOn(column: string) {
    this.sortedColumn$.next(column);
  }

  public getSortDirection(): string {
    return this.activeRequestsService.getSortOrder() ?? defaultSortOrder;
  }
  constructor(
    public state: RxState<SortState>,
    private capabilityRequestService: CapabilityRequestService,
    private capabilityLauncherService: CapabilityLauncherService,
    private pollingDataUpdaterService: PollingDataUpdaterService,
    private dataRetriever: DataRetrieverService,
    private capabilitiesService: CapabilitiesService,
    private route: ActivatedRoute,
    private router: Router,
    private storageService: StorageService,
    private activeRequestsService: ActiveRequestsService,
    private titleService: Title,
  ) {
    // Initialize table sorting to sort on request ID
    this.sortOn("id");
    // Set page title
    this.titleService.setTitle(this.title);

    // Track the current sorting state
    this.state.set(initSortState);
    this.state.connect("sortCol", this.sortedColumn$);
    this.state.connect("sortDir", this.sortDirection$);
  }

  ngOnInit(): void {
    this.qaStaff = [];
    this.getRoutes();

    this.pageSize = this.getPageSizeOptions()[2];
    this.pageStart = 0;
    this.pageStop = this.pageSize;

    // get list of enabled capabilities
    const enabledCapabilitiesObserver = {
      next: (capabilities) => {
        this.enabledCapabilities = capabilities;
      },
      error: (error) => console.error("Error when retrieving capability version:", error),
    };
    this.capabilitiesService.getEnabledCapabilities().subscribe(enabledCapabilitiesObserver);

    this.pollingDataUpdaterService
      .getDataPoller$(this.dataRetriever.getCapabilityUrl(this.capabilityName))
      .pipe(
        takeUntil(this.ngUnsubscribe),
        repeatWhen(() => this.ngUnsubscribe),
      )
      .subscribe(this.capabilityObserver);

    // Set up the observer to receive active requests, this triggers the initial sorting
    this.pollingDataUpdaterService
      .getDataPoller$(
        this.capabilityRequestService.getActiveCapabilityRequestsUrl(
          this.capabilityName,
          this.filters,
        ),
      )
      .pipe(
        takeUntil(this.ngUnsubscribe),
        repeatWhen(() => this.ngUnsubscribe),
      )
      .subscribe(this.activeRequestsObserver);

    const qaStaffObserver = {
      next: (req) => {
        // this will distinguish Stage 1 and Stage 2 reviewers in this.qaStaff
        // like, this.qaStaff["Stage 1"] and this.qaStaff["Stage 2"]
        if (req["Stage 1"]) {
          this.qaStaff["Stage 1"] = req["Stage 1"];
          this.availableFilters.stage_1_reviewer = this.qaStaff["Stage 1"];
          this.checkFilters();
        } else if (req["Stage 2"]) {
          this.qaStaff["Stage 2"] = req["Stage 2"];
          this.availableFilters.stage_2_reviewer = this.qaStaff["Stage 2"];
          this.checkFilters();
        }
      },
      error: (error) => console.error("Error when retrieving QA staff:", error),
    };

    this.capabilitiesService.getQaStaff("Stage 1").subscribe(qaStaffObserver);
    this.capabilitiesService.getQaStaff("Stage 2").subscribe(qaStaffObserver);

    this.setupAvailableFilters();
    this.checkFilters();

    // Get autocomplete suggestion list
    this.filteredStage2QaStaff = this.stage2QaControl.valueChanges.pipe(
      startWith(""),
      map((val) => this.filter(val, "Stage 2")),
    );
    this.filteredStage1QaStaff = this.stage1QaControl.valueChanges.pipe(
      startWith(""),
      map((val) => this.filter(val, "Stage 1")),
    );
  }

  private setupAvailableFilters() {
    this.availableFilters = {
      state: this.statesToFilter,
      exec_status: this.execStatusToFilter,
      download_status: this.downloadStatusToFilter,
      is_srdp: this.srdpOptions,
      obs_min: this.obsDateOptions,
      obs_max: this.obsDateOptions,
    };
  }

  private checkFilters() {
    if (this.filters) {
      Object.entries(this.filters).map(([paramKey, paramValue]) => {
        if (paramKey != "capability") {
          if (Array.isArray(paramValue)) {
            if (this.availableFilters[paramKey]) {
              paramValue.forEach((val) => {
                this.createSelectedFilters(paramKey, val as string);
              });
            }
          } else if (this.availableFilters[paramKey]) {
            this.createSelectedFilters(paramKey, paramValue as string);
          } else if (this.availableFilters[paramKey] === "") {
            this.createSelectedFilters(paramKey, paramValue as string);
          }
        }
      });
    }
  }

  createSelectedFilters(paramKey: string, paramValue: string) {
    const initialFilterSets = this.getSetByFilterType(paramKey);
    const selectedFilters = this.availableFilters[paramKey].find((f) => {
      if (paramKey.includes("obs_")) {
        return true;
      }
      if (f.user_name) {
        return f.user_name == paramValue;
      } else if (f.name) {
        return f.name == paramValue;
      }
    });
    selectedFilters["isChecked"] = true;
    this.showFilterMenu = true;
    this.handleFilterSets({ data: [paramValue], filter: paramKey, event: true }, initialFilterSets);
  }

  public getRoutes() {
    this.route.queryParamMap.subscribe((params) => {
      this.paramsObj = { ...params.keys, ...params };
      // if there are no query parameters, set to std_calibration
      this.capabilityName = this.paramsObj.params["capability"]
        ? this.paramsObj.params["capability"]
        : this.stdCalibrationCapability;
      this.filters = this.paramsObj.params;
    });
  }

  public filterFrontendFilters(activeRequests: any): any {
    let filteredRequests = activeRequests
    if (this.filters && (
        this.filters.hasOwnProperty("is_srdp") ||
        this.filters.hasOwnProperty("obs_min") ||
        this.filters.hasOwnProperty("obs_max") ||
        this.filters.hasOwnProperty("download_status") ||
        this.filters.hasOwnProperty("exec_status"))) {
      // SRDP filter
      if (this.filters.hasOwnProperty("is_srdp")) {
        filteredRequests = filteredRequests.filter((r) => {
          const workingVersion = r.versions[r.versions.length - 1];
          if (workingVersion.parameters.metadata) {
            const isTrue = Array.isArray(this.filters.is_srdp)
              ? this.filters.is_srdp.includes("true")
              : this.filters.is_srdp === "true";
            const isFalse = Array.isArray(this.filters.is_srdp)
              ? this.filters.is_srdp.includes("false")
              : this.filters.is_srdp === "false";
            if (isTrue && isFalse) {
              return (
                workingVersion.parameters.metadata.is_srdp == true ||
                workingVersion.parameters.metadata.is_srdp == false
              );
            } else if (isFalse) {
              return workingVersion.parameters.metadata.is_srdp == false;
            } else if (isTrue) {
              return workingVersion.parameters.metadata.is_srdp == true;
            }
          }
        });
      }
      // Observation Date filter
      if (this.filters.hasOwnProperty("obs_min") || this.filters.hasOwnProperty("obs_max")) {
        const startDate = this.filters["obs_min"] ? new Date(this.filters["obs_min"]) : new Date(0);
        const endDate = this.filters["obs_max"] ? new Date(this.filters["obs_max"]) : new Date();

        filteredRequests = filteredRequests.filter(function (request) {
          const md = request.versions[0].parameters.metadata;
          if (md && md.hasOwnProperty("obs_end_time")) {
            const dateToFilterBy = md.obs_end_time || {};
            const date = new Date(dateToFilterBy);
            return date >= startDate && date <= endDate;
          } else {
            return false;
          }
        });
        // return filteredDates;
      }
      // NGAS Download Status filter
      if (this.filters.hasOwnProperty("download_status")) {
        filteredRequests = filteredRequests.filter(r => {
          return this.filters.download_status.includes(this.getFetchStatus(r));
        });
      }
      // Execution Status filter
      if (this.filters.hasOwnProperty("exec_status")) {
        filteredRequests = filteredRequests.filter(r => {
          return this.filters.exec_status.includes(this.getExecutionStatusName(r));
        });
      }
    }
    return filteredRequests;
  }

  /**
   * Get metadata from capability request parameters if it exists
   * @param request Capability request
   * @return metadata Metadata section of parameters
   */
  public getMetadata(request: CapabilityRequest): any {
    const latestVersion = request.versions[request.versions.length - 1];
    if (latestVersion.parameters.hasOwnProperty("metadata")) {
      return latestVersion.parameters.metadata.valueOf();
    } else {
      return {};
    }
  }

  public togglePauseCapability(): void {
    const togglePauseCapabilityObservable = {
      next: (request) => {},
      error: (error) => {
        console.error("Error when pausing capability: ", error);
      },
    };

    const pauseAction = this.isPaused ? "unpause" : "pause";
    this.dataRetriever
      .togglePauseCapability(this.stdCalibrationCapability, pauseAction)
      .subscribe(togglePauseCapabilityObservable);
  }

  public toggleSRDPCompatibility(request_id: string, isSRDP: boolean): void {
    // stop polling when user clicks checkbox
    this.ngUnsubscribe.next(false);
    const toggleSRDPCompatibilityObservable = {
      next: (request) => {
        // start polling again after SRDP status has changed
        // TODO: success message to user
        this.ngUnsubscribe.next(true);
      },
      error: (error) => {
        console.error("Error when toggling SRDP-Compatibility:", error);
        // start polling again with SRDP status unchanged
        // TODO: error message to user
        this.ngUnsubscribe.next(true);
      },
    };
    this.capabilityRequestService
      .toggleSRDPCompatibility(request_id, isSRDP)
      .subscribe(toggleSRDPCompatibilityObservable);
  }

  public setSelectedCapability(capabilityName: string): void {
    // set capabilityName from dropdown
    this.capabilityName = capabilityName;

    // Store the current capability selection so the header can access it
    this.storageService.saveSession(WsHeaderComponent.previousCapabilityKey, this.capabilityName);

    // update query parameters to new capabilityName
    this.router
      .navigate(["."], {
        relativeTo: this.route,
        queryParams: { capability: this.capabilityName },
      })
      .then(() => {
        // reload to make sure QA specific columns only show for QA-able capabilities
        window.location.reload();
      });

    // Reset paginator to the first page
    this.requestsPaginator.firstPage();
  }

  public setStaff(requestId: string, staffMember: Staff): void {
    // stop polling when user opens dropdown
    this.ngUnsubscribe.next(false);
    const assignQaStaffObservable = {
      next: () => {
        // start polling again when user selects qa staff and req is successful
        // TODO: success message to user
        this.ngUnsubscribe.next(true);
      },
      error: (error) => {
        console.error("Error when setting " + staffMember.group + ", error: ", error);
        // start polling again when user selects qa staff and req fails
        // TODO: error message to user
        this.ngUnsubscribe.next(true);
      },
    };
    this.capabilityRequestService
      .assignQaStaff(requestId, staffMember)
      .subscribe(assignQaStaffObservable);
  }

  // pause polling when user opens qa staff dropdown
  public pausePolling(isDropdownOpen: Event): void {
    isDropdownOpen ? this.ngUnsubscribe.next(false) : this.ngUnsubscribe.next(true);
  }

  public clearStaff(requestId: string, group: string): void {
    return this.setStaff(requestId, {
      user_name: "UNASSIGN",
      group: group,
      available: true,
    });
  }

  // Returns the execution status name based on the current execution state for displaying, sorting, and filtering
  public getExecutionStatusName(request: any): string {
    let statusName = "";
    if (
      request.state !== "Created" &&
      request.versions[request.versions.length - 1].state !== "Created"
    ) {
      statusName = request.versions[request.versions.length - 1].current_execution
        ? request.versions[request.versions.length - 1].current_execution.state_name
        : "";
    }
    if (request.state === 'Created' ||
      (request.state === 'Submitted' && request.versions[request.versions.length - 1].state === 'Created')) {
      statusName = this.notSubmittedState;
    }

    return statusName;
  }

  // Compares 2 values and returns 1, -1, or 0 to determine sorting order
  private compareColumns(a: any, b: any, direction = "asc"): number {
    if (!a || !b) {
      return !a ? 1 : -1;
    } else if (a > b) {
      return direction === "asc" ? 1 : -1;
    } else if (a < b) {
      return direction === "asc" ? -1 : 1;
    } else {
      return 0;
    }
  }

  private sortByColumn(list: any[] | undefined, column: string, direction = "asc"): any[] {
    // Set global sort column and direction flags
    this.sortCol = column;
    this.sortDir = direction;

    // Do sort
    return (list || []).sort((a, b) => {
      if (column.includes("obs_")) {
        // Handle observation times (or really any metadata field)
        a = this.getMetadata(a);
        b = this.getMetadata(b);
      } else if (column.includes("exec_status")) {
        // Handle execution status stored in the latest version
        a = this.getExecutionStatusName(a);
        b = this.getExecutionStatusName(b);
      } else {
        // Handle normally accessible members
        // Make sure empty columns stay at the end of the list for readability
        a = a[column];
        b = b[column];
      }
      return this.compareColumns(a, b, direction);
    });
  }

  toggleFilterMenu() {
    this.showFilterMenu = !this.showFilterMenu;
  }

  emitFilterEvent(e: Filter) {
    const existingQueryParams = this.route.snapshot.queryParams;
    const workingSet = this.getSetByFilterType(e.filter);

    this.handleFilterSets(e, workingSet);
    this.handleNewParams(e, existingQueryParams);

    // Set filters that will be applied when this component refreshes
    this.filters = this.newParams;

    this.router
      .navigate(["."], {
        relativeTo: this.route,
        queryParams: this.newParams,
      })
      .then(() => {
        // TODO: There's probably a better way to refresh this component without a reload
        window.location.reload();
      });
  }

  handleNewParams(filter: Filter, existingQueryParams: any) {
    this.newParams = { ...existingQueryParams };
    if (this.filterSets.hasOwnProperty(filter.filter) && this.filterSets[filter.filter].size == 0) {
      if (this.shouldDeleteFilter(existingQueryParams, filter)) {
        delete this.newParams[filter.filter];
      } else {
        this.newParams[filter.filter] = this.newParams[filter.filter].filter(
          (f) => f !== filter.data[0],
        );
      }
      this.newParams = { ...this.newParams };
    } else {
      this.newParams = {
        ...this.newParams,
        [filter.filter]: [...this.filterSets[filter.filter]],
      };
    }
  }

  shouldDeleteFilter(existingQueryParams: any, filter: Filter): boolean {
    const isLengthIsEqualToOne = Array.isArray(existingQueryParams[filter.filter])
      ? existingQueryParams[filter.filter].length == 1
      : existingQueryParams[filter.filter].size == 1;
    return (
      (existingQueryParams.hasOwnProperty(filter.filter) &&
        isLengthIsEqualToOne &&
        filter.event == false) ||
      !Array.isArray(existingQueryParams[filter.filter])
    );
  }

  getSetByFilterType(s: string): any {
    switch (s) {
      case "state":
        return this.statesSet;
      case "exec_status":
        return this.execStatusSet;
      case "download_status":
        return this.downloadStatusSet;
      case "stage_1_reviewer":
        return this.stage1QaSet;
      case "stage_2_reviewer":
        return this.stage2QaSet;
      case "is_srdp":
        return this.srdpSet;
      case "obs_min":
        return this.obsDateSet;
      case "obs_max":
        return this.obsDateSet;
      default:
        return [];
    }
  }

  handleFilterSets(filter: Filter, setOfFilters: Set<string>) {
    if (filter.filter.includes("obs_")) {
      setOfFilters.clear();
    }
    filter.event ? setOfFilters.add(filter.data[0]) : setOfFilters.delete(filter.data[0]);
    if (!this.filterSets.hasOwnProperty(filter.filter)) {
      this.filterSets[filter.filter] = setOfFilters;
    }
    this.filterSets[filter.filter];
  }

  ngOnDestroy(): void {
    this.ngUnsubscribe.next();
    this.ngUnsubscribe.complete();
  }

  // Filter staff in dropdown text boxes as user inputs text
  filter(input: string, staffMember: string): Staff[] {
    return this.qaStaff[staffMember].filter(
      (option) => option.user_name.toLowerCase().indexOf(input.toLowerCase()) === 0,
    );
  }

  // Control what gets displayed in the text input when selecting a QA staff member
  staffSelectDisplay(staff: Staff): string {
    return "";
  }

  // Handles event when paginator changes page size or moves to next page
  changePage(pageChange: PageEvent): void {
    this.pageStart = pageChange.pageIndex * pageChange.pageSize;
    this.pageStop = this.pageStart + pageChange.pageSize;
  }

  trackActiveRequests(index: number, activeRequest: any) {
    return activeRequest.id;
  }

  getFetchStatus(request: CapabilityRequest): string {
    // figure out the latest version
    if (request.versions.length == 0)
      return "Not Started";

    let version = request.versions[request.versions.length - 1];

    if (version.workflow_metadata && version.workflow_metadata.fetch_end_time)
      return "Complete";
    else if (version.workflow_metadata && version.workflow_metadata.fetch_start_time)
      return "In Progress";
    else
      return "Unknown";
  }
}
