#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Notifications manager """

from typing import List

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

from workspaces.notification.schema import NotificationTemplate
from workspaces.notification.services.interfaces import NotificationInfoIF

Base = declarative_base()


class NotificationInfo(NotificationInfoIF):
    """Manages notifications"""

    def __init__(self, session: Session):
        self.session = session

    def all_templates(self) -> List[NotificationTemplate]:
        return self.session.query(NotificationTemplate).order_by(NotificationTemplate.name.desc()).all()

    def lookup_template(self, template_name: str) -> NotificationTemplate:
        return self.session.query(NotificationTemplate).filter_by(name=template_name).first()

    def create_template(self, template_name: str, description: str, content: str) -> NotificationTemplate:
        """
        Create new notification template and save it in the database

        :param template_name: Name of new template
        :param description: brief description of template
        :param template: the full text of the template
        :return: Created NotificationTemplate
        """
        template = NotificationTemplate(name=template_name, description=description, content=content)
        self.save_entity(template)
        return template

    def delete_template(self, template_name: str):
        template = self.session.query(NotificationTemplate).filter_by(name=template_name).first()
        self.session.delete(template)

    def update_template(self, template_name: str, template_text: str):
        """
        Replace the content of a notification template.

        :param template_name: Name of template
        :param template_text: full text of the template
        :return:
        """
        template = self.lookup_template(template_name)

        template.content = template_text
        self.save_entity(template)

    def save_entity(self, entity: Base):
        """
        Save entity to database

        :param entity:
        """
        self.session.add(entity)
        self.session.flush()
